@@ -13,13 +13,13 @@ import subprocess
 from mash import io_util
 from mash.filesystem.filesystem import FileSystem
 from mash.io_util import log, shell_ready_signal, print_shell_ready_signal, check_output
-from mash.util import for_any, has_method, identity, is_alpha, is_globbable, is_valid_method_name, match_words, omit_prefixes, split_prefixes, split_sequence, glob
+from mash.shell import delimiters
+from mash.shell.delimiters import DEFINE_FUNCTION, LEFT_ASSIGNMENT, RIGHT_ASSIGNMENT
 from mash.shell.function import InlineFunction
+from mash.util import for_any, has_method, identity, is_globbable, is_valid_method_name, match_words, omit_prefixes, split_prefixes, split_sequence, glob
+
 
 confirmation_mode = False
-bash_delimiters = ['|', '>', '>>', '1>', '1>>', '2>', '2>>']
-py_delimiters = [';', '|>', '>>=']
-other_delimiters = ['=', '<-', '->']
 default_session_filename = '.shell_session.json'
 
 
@@ -94,16 +94,13 @@ class BaseShell(Cmd):
     def delimiters(self):
         """Return the most recent values of the delimiters.
         """
-        delimiters = py_delimiters + bash_delimiters + ['->']
-
-        # insert '<-' after ';'
-        delimiters.insert(1, '<-')
-
-        return delimiters
+        items = delimiters.python + delimiters.bash
+        items.remove('=')
+        return items
 
     def set_infix_operators(self):
         # use this for infix operators, e.g. `a = 1`
-        self.infix_operators = {'=': self.handle_equation}
+        self.infix_operators = {'=': self.handle_set_env_variable}
         # the sign to indicate that a variable should be expanded
         self.variable_prefix = '$'
 
@@ -114,22 +111,12 @@ class BaseShell(Cmd):
 
         Naming conflicts with existing `delimiters` are resolved.
         """
+        for char in chars:
+            if char in delimiters.all:
+                raise ShellError(f'Char {char} is already in use.')
+
         self._do_char_method = method
         self._chars_allowed_for_char_method = chars
-        self.resolve_char_name_conflicts()
-
-    def resolve_char_name_conflicts(self):
-        # TODO don't mutate global vars
-        for char in self._chars_allowed_for_char_method:
-            if char in bash_delimiters:
-                logging.warning(
-                    f'Overriding default sh delimiters: remove {char}')
-                bash_delimiters.remove(char)
-
-            if char in py_delimiters:
-                logging.warning(
-                    f'Overriding default py delimiters: remove {char}')
-                py_delimiters.remove(char)
 
     def update_env(self, env: Dict[str, Any] = None):
         if env is None:
@@ -141,14 +128,16 @@ class BaseShell(Cmd):
 
         self.env = env
 
-    def eval(self, args: Iterable[str]) -> Types:
+    def eval(self, args: Iterable[str], quote=True) -> Types:
         """Evaluate / run `args` and return the result.
         """
-        # convert args to a shell command
+        if quote:
+            args = (shlex.quote(arg) for arg in args)
+            # args = ' '.join(shlex.quote(arg) for arg in args)
+
+        args = ' '.join(args)
 
         k = '_eval_output'
-
-        args = ' '.join(shlex.quote(arg) for arg in args)
         line = f'{args} |> export {k}'
 
         self.onecmd(line)
@@ -173,18 +162,40 @@ class BaseShell(Cmd):
 
         raise RuntimeError('Cannot retrieve result')
 
-    def handle_equation(self, lhs: Tuple[str], *rhs: str) -> str:
-        if len(lhs) == 1:
-            k = lhs[0]
-            self.set_env_variable(k, *rhs)
-            return ''
+    def handle_set_env_variable(self, lhs: Tuple[str], *rhs: str) -> str:
+        if len(lhs) != 1:
+            raise ShellError()
 
-        f, *args = lhs
-        implementation = ' '.join(rhs[1:])
-        self.add_inline_function(f, args, implementation)
+        k = lhs[0]
+        self.set_env_variable(k, *rhs)
         return ''
 
-    def add_inline_function(self, f, args, inner) -> str:
+    def handle_define_inline_function(self, terms: List[str]) -> str:
+        f, *args = terms
+        args = [arg for arg in args if arg != '']
+
+        if not args or not args[0].startswith('(') or not args[-1].endswith(')'):
+            lhs = ' '.join(terms)
+            raise ShellError(f'Invalid syntax for inline function: {lhs}')
+
+        # strip braces
+        if args[0] == '()':
+            args = []
+        else:
+            args[0] = args[0][1:]
+            args[-1] = args[-1][:-1]
+
+        self.locals.set(DEFINE_FUNCTION,
+                        InlineFunction('', *args, func_name=f))
+
+    def _save_inline_function(self) -> str:
+        func = self.locals[DEFINE_FUNCTION]
+        self.locals.rm(DEFINE_FUNCTION)
+
+        f = func.func_name
+        args = func.args
+        inner = func.command
+
         if has_method(self, f'do_{f}'):
             raise ShellError(
                 f'Name conflict: Cannot define inline function {f}, '
@@ -204,19 +215,23 @@ class BaseShell(Cmd):
     def call_inline_function(self, f: InlineFunction, *args: str):
         translations = {}
 
-        for i, k in enumerate(f.args):
-            translations[k] = args[i]
+        if len(args) != len(f.args):
+            raise ShellError(
+                f'Invalid number of arguments: {len(f.args)} arguments expected .')
 
-        terms = list(self.translate_terms(f.command.split(' '), translations))
-        line = ' '.join(terms)
+        for i, k in enumerate(f.args):
+            # quote item to preserve `\n`
+            translations[k] = shlex.quote(args[i])
+
+        terms = [term for term in f.command.split(' ') if term != '']
+        terms = list(self.translate_terms(terms, translations))
 
         first_func = terms[0]
         if not has_method(self, f'do_{first_func}') \
                 and first_func not in self.locals['functions']:
             terms = ['print'] + terms
 
-        line = ' '.join(terms)
-        return super().onecmd(line)
+        return self.eval(terms, quote=False)
 
     def set_env_variable(self, k: str, *values: str):
         """Set the variable `k` to `values`
@@ -228,6 +243,18 @@ class BaseShell(Cmd):
         self.env[k] = ' '.join(values)
         return k
 
+    def set_env_variables(self, keys: str, result: str):
+        """Set the variable `k` to `values`
+        """
+        if result is None:
+            raise ShellError(f'Missing return value in assignment: {keys}')
+
+        # TODO set multiple keys based on pattern matching
+        # e.g. i j = range(2)
+        k = keys[0]
+
+        self.env[k] = result
+
     def show_env(self, env=None):
         if env is None:
             env = self.env
@@ -309,12 +336,23 @@ class BaseShell(Cmd):
         # results in a = 10
         ```
         """
-        k = args.split(' ')[0]
+        keys = args.split(' ')
 
-        if not is_valid_method_name(k):
-            raise ShellError('Invalid variable name format: {k}')
+        for k in keys:
+            if not is_valid_method_name(k):
+                raise ShellError('Invalid variable name format: {k}')
 
-        self.locals.set('assignee', k)
+        if LEFT_ASSIGNMENT in self.locals:
+            assignee = self.locals[LEFT_ASSIGNMENT]
+
+            # cancel previous assignments
+            self.locals.rm(LEFT_ASSIGNMENT)
+
+            raise ShellError(
+                f'Assignments cannot be used inside other assignments: {assignee}')
+
+        self.locals.set(LEFT_ASSIGNMENT, keys)
+        self.set_env_variables(keys, '')
 
         # return value must be empty to prevent side-effects in the next command
         return ''
@@ -326,7 +364,7 @@ class BaseShell(Cmd):
         if not args:
             return ''
 
-        k, *values = args.split()
+        k, *values = args.split(' ')
 
         if len(values) == 0:
             log(f'unset {k}')
@@ -344,6 +382,58 @@ class BaseShell(Cmd):
         logging.info(f'Cmd = !{args}')
         return check_output(args)
 
+    def do_reduce(self, *args: str):
+        """Reduce a sequence of items to using an operator.
+
+        See https://en.wikipedia.org/wiki/Reduction_operator
+
+        E.g. compute the sum:
+        `range 10 |> reduce sum 0 $
+        """
+        return self.do_foldr(*args)
+
+    def do_foldr(self, args: str, delimiter='\n'):
+        """Fold or reduce from right to left.
+
+        See https://wiki.haskell.org/Foldr_Foldl_Foldl'
+        """
+        lines = args.split(delimiter)
+        msg = 'Not enough arguments. Usage: `foldl f zero [args] $ [args]`.'
+        # if len(lines) <= 1:
+        if not lines:
+            log(msg)
+            return
+
+        items = lines[0].split(' ')
+        if len(items) <= 1:
+            log(msg)
+            return
+
+        f, zero, *args, line = items
+        lines = [line] + lines[1:]
+
+        if '$' in args:
+            i = args.index('$')
+        else:
+            i = -1
+
+        # apply the reduction
+        acc = zero
+        for line in lines:
+            local_args = args.copy()
+            line = line.split(' ')
+
+            if i == -1:
+                local_args += line
+            else:
+                local_args[i:i+1] = line
+
+            line = [f, acc] + local_args
+
+            acc = self.run_single_command(line)
+
+        return acc
+
     def do_map(self, args: str, delimiter='\n'):
         """Apply a function to every line.
         If `$` is present, then each line from stdin is inserted there. 
@@ -357,7 +447,7 @@ class BaseShell(Cmd):
         ```
         """
         lines = args.split(delimiter)
-        msg = 'Not enough arguments. Usage: `map f [args] *`.'
+        msg = 'Not enough arguments. Usage: `map f [args] $ [args]`.'
         if len(lines) <= 1:
             log(msg)
             return
@@ -433,10 +523,6 @@ class BaseShell(Cmd):
         try:
             line = self.onecmd_prehook(line)
             lines = list(self.parse_commands(line))
-
-            if any(line[0] == '<-' for line in lines):
-                lines[0] = ['assign'] + lines[0]
-
             self.run_commands(lines)
 
         except CancelledError:
@@ -490,7 +576,24 @@ class BaseShell(Cmd):
         if not lines:
             return
 
-        for line in lines:
+        for i, line in enumerate(lines):
+
+            if LEFT_ASSIGNMENT not in self.locals \
+                    and i+1 < len(lines) \
+                    and '<-' in lines[i+1]:
+                # prefix line if '<-' is used later on
+                j = 1 if ';' in line else 0
+                line.insert(j, 'assign')
+
+            if DEFINE_FUNCTION in self.locals:
+                if ':' in line:
+                    line = line[1:]
+
+                # self.locals[DEFINE_FUNCTION].command += '; ' + ' '.join(line)
+                self.locals[DEFINE_FUNCTION].command += ' ' + ' '.join(line)
+                result = None
+                continue
+
             try:
                 result = self.run_single_command(line, result)
 
@@ -510,9 +613,18 @@ class BaseShell(Cmd):
 
                 raise ShellError(str(e))
 
-        if 'assignee' in self.locals:
+            if DEFINE_FUNCTION not in self.locals:
+                # handle inline `<-`
+                if LEFT_ASSIGNMENT in self.locals and 'assign' not in line:
+                    self._save_assignee(result)
+                    result = ''
+
+        if DEFINE_FUNCTION in self.locals:
+            self._save_inline_function()
+
+        elif LEFT_ASSIGNMENT in self.locals and not io_util.interactive:
+            # cancel assignment
             self._save_assignee(result)
-            result = ''
 
         elif result is not None:
             print(result)
@@ -524,7 +636,7 @@ class BaseShell(Cmd):
             command_and_args)
 
         if prefixes:
-            if prefixes[-1] in bash_delimiters:
+            if prefixes[-1] in delimiters.bash:
                 return self.pipe_cmd_sh(line, result, delimiter=prefixes[-1])
 
             elif prefixes[-1] == '>>=':
@@ -533,34 +645,35 @@ class BaseShell(Cmd):
                 line = f'map {line}'
                 return self.pipe_cmd_py(line, result)
 
-            elif prefixes[-1] == '->':
-                # TODO verify syntax
+            elif prefixes[-1] == RIGHT_ASSIGNMENT:
+                # TODO verify syntax of `line`
                 assert ' ' not in line
-                self.set_env_variable(line, result)
+                self.set_env_variables(line, result)
                 return ''
 
         if infix_operator_args:
             return self.infix_command(*infix_operator_args)
+        elif is_function_definition(command_and_args):
+            return self.handle_define_inline_function(command_and_args)
 
         return self.pipe_cmd_py(line, result)
 
     def _save_assignee(self, result: str):
-        k = self.locals['assignee']
+        keys = self.locals[LEFT_ASSIGNMENT]
+
+        self.locals.rm(LEFT_ASSIGNMENT)
 
         if result is None:
-            raise ShellError(f'Missing return value in assignment: {k}')
+            raise ShellError(f'Missing return value in assignment: {keys}')
         elif result == '' and self._last_results:
             result = self._last_results.pop()
 
-        self.env[k] = result
-        self.locals.rm('assignee')
+        self.set_env_variables(keys, result)
 
     def filter_result(self, command_and_args, result):
         if ';' in command_and_args:
-            if 'assignee' in self.locals:
-                self._save_assignee(result)
 
-            elif result is not None:
+            if result is not None and LEFT_ASSIGNMENT not in self.locals:
                 # print prev result & discard it
                 print(result)
 
@@ -578,7 +691,7 @@ class BaseShell(Cmd):
         prefixes = list(split_prefixes(command_and_args, self.delimiters))
         prefix = prefixes[-1]
 
-        if prefix in bash_delimiters:
+        if prefix in delimiters.bash:
             return prefix
 
     def parse_single_command(self, command_and_args: List[str]) -> Tuple[List[str], str, List[str]]:
@@ -588,7 +701,6 @@ class BaseShell(Cmd):
         args = list(self.expand_variables(args))
         line = ' '.join(chain.from_iterable(([f], args)))
 
-        # TODO make this check quote-aware
         there_is_an_infix_operator = for_any(
             self.infix_operators, contains, args)
 
@@ -609,7 +721,7 @@ class BaseShell(Cmd):
         """
         May raise subprocess.CalledProcessError
         """
-        assert delimiter in bash_delimiters
+        assert delimiter in delimiters.bash
 
         # pass last result to stdin
         line = f'echo {shlex.quote(prev_result)} {delimiter} {line}'
@@ -639,6 +751,24 @@ class BaseShell(Cmd):
             # e.g. convert 'echo "echo 1"' to ['echo', 'echo 1']
             terms = shlex.split(line, comments=True)
 
+            # re-quote delimiters
+            for i, term in enumerate(terms):
+                # if other_delimiters:
+                if term in self.delimiters or term == '=':
+                    if '"' + term + '"' in line or "'" + term + "'" in line:
+                        terms[i] = f'"{terms[i]}"'
+
+            # split `:`
+            for i, term in enumerate(terms):
+                if term.endswith(':'):
+                    # verify that the term wasn't quoted
+                    if '"' + term + '"' in line or "'" + term + "'" in line:
+                        continue
+
+                    terms[i] = terms[i][:-1]
+                    terms.insert(i+1, ':')
+                    break
+
         except ValueError as e:
             msg = f'Invalid syntax: {e} for {str(line)[:10]} ..'
             if self.ignore_invalid_syntax:
@@ -767,3 +897,19 @@ class BaseShell(Cmd):
                 continue
 
             yield term
+
+
+def is_function_definition(terms: List[str]) -> bool:
+    terms = [term for term in terms if term != '']
+
+    if len(terms) < 2:
+        return
+
+    if len(terms) == 2:
+        term = terms[-1]
+        first = term
+        last = term
+    else:
+        _f, first, *_, last = terms
+
+    return first.startswith('(') and last.endswith(')')
